import { ChartD3Service } from '../../services/chart.d3.service';
import { ElementStrategyFactory, SingleLineSeries } from './chart.service';
import { SeriesType } from '../../interfaces/SeriesType';
import { ChartEvents } from './chart.events';
export class Chart {
    constructor() {
        this.service = new SingleLineSeries();
    }
    componentWillLoad() {
        this.axisConfigs = {
            x: this.service.getDefaultXAxisConfig(),
            y: this.service.getDefaultYAxisConfig()
        };
    }
    componentDidLoad() {
        this.debounceCreation();
    }
    async onColumnsLoaded() {
        this.columns = await this.loadColumns();
        this.debounceCreation();
    }
    async onLinesLoaded() {
        this.lines = await this.loadLines();
        this.debounceCreation();
    }
    async onStackedColumnsLoaded() {
        this.stackedColumns = await this.loadStackedColumns();
        this.debounceCreation();
    }
    async chartComponentsUpdate() {
        this.popoverComponent = await this.loadPopover();
    }
    async onXAxisLoaded({ detail }) {
        this.xAxisElement = await this.loadXAxis();
        this.axisConfigs.x = Object.assign({}, this.axisConfigs.x, detail);
        this.debounceCreation();
    }
    async onYAxisLoaded({ detail }) {
        this.yAxisElement = await this.loadYAxis();
        this.axisConfigs.y = Object.assign({}, this.axisConfigs.y, detail);
        this.debounceCreation();
    }
    handleResize() {
        this.debounceCreation();
    }
    addLineEvents(svg, chartOverlay, x, data) {
        chartOverlay
            .on('mousemove', () => this.eventsService.anjHoverHandler(svg, data, x, this.axisConfigs, this.seriesMeta))
            .on('mouseout', () => this.eventsService.anjEndHoverHandler(svg));
    }
    addColumnEvents(svgBars, data) {
        svgBars
            .on('mousemove', (_d, i, bars) => this.eventsService.anjColumnHoverHandler(bars[i], data, i, this.axisConfigs, this.seriesMeta))
            .on('mouseout', () => this.eventsService.anjEndHoverHandler(svgBars));
    }
    addStackedEvents(svg, bars) {
        bars
            .on('mousemove', (_d, i) => this.eventsService.anjStackedHoverHandler(svg, this.popoverData, i, this.axisConfigs, this.seriesMeta))
            .on('mouseout', () => this.eventsService.anjEndHoverHandler(svg));
    }
    appendSVG(svg) {
        while (this.chartContainer.children.length > 1) {
            this.chartContainer.removeChild(this.chartContainer.lastChild);
        }
        this.chartContainer.appendChild(svg.node());
    }
    debounceCreation() {
        window.clearTimeout(this.creationTimer);
        this.creationTimer = window.setTimeout(async () => {
            this.chartSVG = await this.create();
        }, Chart.creationDebounceLength);
    }
    async create() {
        const dims = this.service.getDims(this.chartContainer);
        this.primarySeries = (this.stackedColumns) ? SeriesType.Stacked : (this.lines && !this.columns) ? SeriesType.Line : SeriesType.Column;
        this.secondarySeries = ((this.stackedColumns || this.columns) && this.lines) ? SeriesType.Line : undefined;
        this.primaryComponents = (this.stackedColumns) ? this.stackedColumns : (this.lines && !this.columns) ? this.lines : this.columns;
        this.secondaryComponents = ((this.stackedColumns || this.columns) && this.lines) ? this.lines : undefined;
        this.service = ElementStrategyFactory.get(this.primarySeries, this.primaryComponents);
        this.multiSeries = this.service.isMultiSeries();
        this.popover.placement = this.service.getPopoverPlacement();
        this.axisConfigs = this.service.xAxisTypeOverride(this.axisConfigs);
        this.eventsService = new ChartEvents(this.primarySeries, this.anjHover, this.service, this.popover);
        const { data, stackedData, minMax } = this.service.formatPrimaryData(this.axisConfigs);
        const flattenedData = this.service.flatten(data);
        this.popoverData = data;
        this.stackedData = stackedData;
        const svg = ChartD3Service.createSVG(dims);
        const x = (this.xAxisElement)
            ? await this.hydrateXAxis(dims, flattenedData, svg, this.primarySeries)
            : ChartD3Service.addXAxis(svg, dims, flattenedData, this.axisConfigs, undefined, this.primarySeries);
        const y = (this.yAxisElement)
            ? await this.hydrateYAxis(dims, flattenedData, svg, minMax)
            : ChartD3Service.addYAxis(svg, dims, flattenedData, this.axisConfigs, minMax);
        await this.hydratePrimaryComponents(svg, x, y, data, dims);
        await this.hydrateSecondaryComponents(svg, x, y, dims);
        this.appendSVG(svg);
        return Promise.resolve(svg);
    }
    async hydratePrimaryComponents(svg, x, y, data, dims) {
        switch (this.primarySeries) {
            case SeriesType.Column:
                const x1 = ChartD3Service.getX1(x, data, this.multiSeries);
                return this.hydrateColumns(svg, data, x, x1, y, dims);
            case SeriesType.Stacked:
                return this.hydrateStackedColumns(svg, this.stackedData, x, y, dims);
            case SeriesType.Line:
                return this.hydrateLinesPrimary(svg, data, x, y, dims);
            default:
                return Promise.resolve();
        }
    }
    async hydrateSecondaryComponents(svg, x, y, dims) {
        switch (this.secondarySeries) {
            case SeriesType.Line:
                const lineData = this.secondaryComponents.map(l => l.data);
                const formattedLineData = this.service.coreFormatData(this.axisConfigs, lineData);
                await this.hydrateLinesSecondary(svg, formattedLineData, x, y, dims);
            default:
                return Promise.resolve();
        }
    }
    async hydrateColumns(svg, data, x, x1, y, dims) {
        this.seriesMeta = [];
        this.columns.forEach(async (column, i) => {
            const bars = await column.init(svg, x, x1, y, data[i], i, dims);
            this.seriesMeta[i] = { name: column.name, netNegative: column.netNegative, color: undefined };
            this.addColumnEvents(bars, data[i]);
        });
    }
    async hydrateLinesPrimary(svg, data, x, y, dims) {
        this.seriesMeta = [];
        this.lines.forEach(async (line, i) => {
            const color = await line.init(svg, x, y, data[i], i, dims, this.service.isMultiSeries());
            this.seriesMeta[i] = { name: line.name, color, netNegative: line.netNegative };
        });
        const chartOverlay = ChartD3Service.addOverlay(svg, dims);
        this.addLineEvents(svg, chartOverlay, x, data);
        ChartD3Service.addRevealAnimation(svg, dims, this.axisConfigs);
    }
    async hydrateLinesSecondary(svg, data, x, y, dims) {
        this.lines.forEach(async (line, i) => {
            this.popoverData.push(data[i]);
            const color = await line.init(svg, x, y, data[i], i, dims, this.service.isMultiSeries(), this.primaryComponents.length);
            const popoverIndex = i + this.seriesMeta.length;
            this.seriesMeta[popoverIndex] = { name: line.name, color, netNegative: line.netNegative };
        });
    }
    async hydrateStackedColumns(svg, data, x, y, dims) {
        this.seriesMeta = [];
        this.stackedColumns.forEach(async (column, i) => {
            const { bars, color } = await column.init(svg, x, y, data[i], i, dims);
            this.seriesMeta[i] = { name: column.name, netNegative: column.netNegative, color };
            this.addStackedEvents(svg, bars);
        });
    }
    hydrateXAxis(dims, data, svg, primarySeries) {
        return this.xAxisElement.init(svg, dims, data, this.axisConfigs, undefined, primarySeries);
    }
    hydrateYAxis(dims, data, svg, minMax) {
        return this.yAxisElement.init(svg, dims, data, this.axisConfigs, minMax, undefined);
    }
    loadColumns() {
        return Promise.all(Array.from(this.chartContainer.querySelectorAll('anj-column-series')).map(col => col.componentOnReady()));
    }
    loadLines() {
        return Promise.all(Array.from(this.chartContainer.querySelectorAll('anj-line-series')).map(col => col.componentOnReady()));
    }
    loadStackedColumns() {
        return Promise.all(Array.from(this.chartContainer.querySelectorAll('anj-stacked-column-series')).map(col => col.componentOnReady()));
    }
    loadPopover() {
        return Promise.resolve(this.chartContainer.querySelector('anj-popover-content-single, anj-popover-content-multi').componentOnReady());
    }
    loadXAxis() {
        return Promise.resolve(this.chartContainer.querySelector('anj-axis[dimension="x"]').componentOnReady());
    }
    loadYAxis() {
        return Promise.resolve(this.chartContainer.querySelector('anj-axis[dimension="y"]').componentOnReady());
    }
    hostData() {
        return {
            class: {
                'anj-rm-pop-arrow': !this.service.arrow()
            }
        };
    }
    render() {
        return [
            h("anj-overlay", { class: "popover", placement: "top", offset: "0, 15", ref: r => this.popover = r }, (this.popoverComponent) ? [h("anj-arrow", null), h("slot", null)] : undefined)
        ];
    }
    static get is() { return "anj-chart"; }
    static get properties() { return {
        "chartContainer": {
            "elementRef": true
        },
        "multiSeries": {
            "state": true
        },
        "popoverComponent": {
            "state": true
        }
    }; }
    static get events() { return [{
            "name": "anjHover",
            "method": "anjHover",
            "bubbles": true,
            "cancelable": true,
            "composed": true
        }]; }
    static get listeners() { return [{
            "name": "anjColumnDidLoad",
            "method": "onColumnsLoaded"
        }, {
            "name": "anjLineDidLoad",
            "method": "onLinesLoaded"
        }, {
            "name": "anjStackedColumnDidLoad",
            "method": "onStackedColumnsLoaded"
        }, {
            "name": "anjPopoverDidLoad",
            "method": "chartComponentsUpdate"
        }, {
            "name": "anjXAxisDidLoad",
            "method": "onXAxisLoaded"
        }, {
            "name": "anjYAxisDidLoad",
            "method": "onYAxisLoaded"
        }, {
            "name": "window:resize",
            "method": "handleResize",
            "passive": true
        }]; }
    static get style() { return "/**style-placeholder:anj-chart:**/"; }
}
Chart.creationDebounceLength = 100;
