import { d3 } from '../d3-bundle';
import { ChartDataService } from './chart.data.service';
import { SeriesType } from '../interfaces/SeriesType';
export class ChartD3Service {
    static addArea(svg, line, y, data, color) {
        const area = d3.area()
            .curve(d3.curveMonotoneX)
            .defined(line.defined())
            .x(line.x())
            .y1(line.y())
            .y0(y(0));
        svg.append('path')
            .data([data])
            .attr('class', 'series area')
            .attr('fill', color)
            .attr('d', area);
        return area;
    }
    static addStackedColumn(svg, x, y, data, color) {
        const seriesContainers = svg
            .append('g')
            .attr('class', 'series-container')
            .selectAll('.series-container')
            .data([data])
            .enter().append('g')
            .attr('class', 'group')
            .style('fill', color);
        const chartBars = seriesContainers
            .selectAll('rect')
            .data((d) => d)
            .enter().append('rect');
        chartBars
            .attr('class', (d, i) => {
            return d[0] < 0 ? `bar negative bar-${i}` : `bar positive bar-${i}`;
        })
            .attr('width', x.bandwidth() - ChartD3Service.groupPadding)
            .attr('x', (d) => x(d.data.categoryValue) + (ChartD3Service.groupPadding / 2))
            .attr('y', y(0))
            .attr('height', 0);
        chartBars
            .transition()
            .duration(ChartD3Service.animationLength)
            .attr('y', (d) => y(d[1]))
            .attr('height', (d) => y(d[0]) - y(d[1]));
        return chartBars;
    }
    static addColumn(svg, x, x1, y, data, color, _dims, index) {
        const seriesContainers = svg
            .append('g')
            .attr('class', 'series-container')
            .selectAll('.series-container')
            .data([data])
            .enter().append('g');
        const chartBars = seriesContainers
            .attr('class', 'series')
            .attr('transform', 'translate(' + x1(index) + ',0)')
            .style('fill', color)
            .selectAll('rect')
            .data((d) => d)
            .enter().append('rect');
        chartBars
            .attr('class', (d) => {
            return d.y < 0 ? 'bar negative' : 'bar positive';
        })
            .style('opacity', (d) => {
            return d.y < 0 ? .6 : 1;
        })
            .attr('width', x1.bandwidth() - ChartD3Service.groupPadding)
            .attr('x', (d) => x(d.x) + (ChartD3Service.groupPadding / 2))
            .attr('y', () => y(0))
            .attr('height', 0);
        chartBars
            .transition()
            .duration(ChartD3Service.animationLength)
            .attr('y', (d) => d.y >= 0 ? (y(d.y)) : y(0))
            .attr('height', (d) => {
            return Math.max(Math.abs(y(d.y) - y(0)), 1);
        });
        return chartBars;
    }
    static addLine(svg, x, y, data, color) {
        let xOffset = 0;
        if (x.bandwidth) {
            xOffset = ((x.bandwidth()) / 2);
        }
        const valueLine = d3.line()
            .curve(d3.curveMonotoneX)
            .defined((d) => d.y !== null)
            .x((d) => x(d.x) + xOffset)
            .y((d) => y(d.y));
        svg.append('path')
            .data([data])
            .attr('class', 'line')
            .attr('stroke', color)
            .attr('d', valueLine);
        return valueLine;
    }
    static addOverlay(svg, dims) {
        const overlay = svg.append('rect')
            .attr('class', 'overlay')
            .attr('width', dims.width)
            .attr('height', dims.height);
        return overlay;
    }
    static addPoints(svg, x, y, data, color, index) {
        let xOffset = 0;
        if (x.bandwidth) {
            xOffset = ((x.bandwidth()) / 2);
        }
        svg.selectAll(`.point.series${index}`)
            .data(data)
            .enter()
            .append('svg:circle')
            .style('opacity', 0)
            .style('stroke', color)
            .attr('class', (_d, i) => `point point-${i}`)
            .attr('r', () => 1)
            .attr('cx', (d) => x(d.x) + xOffset)
            .attr('cy', (d) => y(d.y));
    }
    static addRevealAnimation(svg, dims, axisConfigs) {
        const delay = ChartD3Service.focusLength;
        const duration = ChartD3Service.animationLength;
        let curtainHeight = dims.height;
        let curtainWidth = dims.width;
        if (!axisConfigs.x.showAxis) {
            curtainHeight = curtainHeight + 1;
        }
        if (axisConfigs.y.showAxis) {
            curtainWidth = curtainWidth - (this.yPad * 2.5);
        }
        const curtain = svg.append('rect')
            .attr('x', -1 * dims.width)
            .attr('y', -1 * curtainHeight)
            .attr('height', curtainHeight)
            .attr('width', curtainWidth)
            .attr('class', 'curtain')
            .attr('transform', 'rotate(180)')
            .style('fill', '#ffffff');
        const t = svg.transition()
            .delay(delay)
            .duration(duration);
        t.select('rect.curtain')
            .attr('width', 0);
        t.select('line.guide')
            .attr('transform', 'translate(' + dims.width + ', 0)');
        return curtain;
    }
    static addStackedScrubbers(svg, x, data, dims, index) {
        const xOffset = ((x.bandwidth() + ChartD3Service.calculatePadding(data, dims)) / 2);
        svg.selectAll(`.scrubber.series${index}`)
            .data(data)
            .enter()
            .append('line')
            .attr('x1', (d) => x(d.data.categoryValue) + xOffset)
            .attr('y1', 0)
            .attr('x2', (d) => x(d.data.categoryValue) + xOffset)
            .attr('y2', dims.height)
            .attr('class', (_d, i) => `scrubber scrubber-${i}`)
            .attr('stroke', '#ffffff')
            .style('opacity', '0');
    }
    static addScrubbers(svg, x, data, dims, index) {
        svg.selectAll(`.scrubber.series${index}`)
            .data(data)
            .enter()
            .append('line')
            .attr('x1', (d) => x(d.x))
            .attr('y1', 0)
            .attr('x2', (d) => x(d.x))
            .attr('y2', dims.height)
            .attr('class', (_d, i) => `scrubber scrubber-${i}`);
    }
    static getX1(x, data, multiSeries) {
        const seriesCount = (multiSeries) ? data.length : 1;
        return d3.scaleBand()
            .rangeRound([0, x.bandwidth()])
            .domain(d3.range(seriesCount))
            .padding(0);
    }
    static addXAxis(svg, dims, data, axisConfigs, _minMax, primarySeries) {
        const x = ChartD3Service.axisTyper(primarySeries, data, dims)
            .range([this.calculateLeftPad(dims, axisConfigs, data), dims.width]);
        const uniqueX = Array.from(new Set(data.map(item => item.x.toString())));
        const availableSlots = (dims.width / 60);
        const tickModulo = Math.max(Math.floor(uniqueX.length / availableSlots), 1);
        const xAxis = (primarySeries === SeriesType.Line)
            ? d3.axisBottom(x)
            : d3.axisBottom(x).tickValues(x.domain().filter((_d, i) => !(i % tickModulo)));
        xAxis.tickSizeOuter(0)
            .tickFormat(ChartDataService.xFormatter.bind(this, ChartDataService.calculateInterval(data), axisConfigs.x))
            .scale(x);
        svg.append('g')
            .attr('class', 'x axis')
            .attr('transform', `translate(0,${dims.height})`)
            .call(xAxis);
        if (!axisConfigs.x.showAxis) {
            svg.select('.x.axis').style('display', 'none');
        }
        if (!axisConfigs.x.showTicks) {
            svg.selectAll('.x.axis .tick').style('display', 'none');
        }
        return x;
    }
    static axisTyper(type, data, dims) {
        switch (type) {
            case SeriesType.Line:
                return d3.scaleTime()
                    .domain(ChartDataService.determineDateRange(data));
            default:
                const domain = data.map((d) => d.x.toString());
                return d3.scaleBand()
                    .padding(ChartD3Service.calculatePadding(data, dims))
                    .domain(domain);
        }
    }
    static calculatePadding(data, dims) {
        const colWidth = dims.width / (data.length - 1);
        return colWidth > 30 ? .05 : 0;
    }
    static addYAxis(svg, dims, data, axisConfigs, minMax) {
        const domain = (minMax) ? [minMax.min, minMax.max] : [0, d3.max(data, (d) => d.y)];
        const y = d3.scaleLinear()
            .domain(domain)
            .nice()
            .range([dims.height, dims.margin.top]);
        const yAxis = d3.axisLeft(y)
            .tickSize(0)
            .ticks(this.calculateTicks(dims.height))
            .tickFormat(ChartDataService.yFormatter.bind(this, axisConfigs.y))
            .scale(y);
        svg.append('g')
            .attr('class', 'y axis')
            .attr('transform', `translate(${this.calculateLeftPad(dims, axisConfigs, data)},0)`)
            .call(yAxis);
        if (!axisConfigs.y.showAxis) {
            svg.select('.y.axis').style('display', 'none');
        }
        if (!axisConfigs.y.showTicks) {
            svg.selectAll('.y.axis .tick').style('display', 'none');
        }
        return y;
    }
    static createSVG(dims) {
        const svg = d3.create('svg');
        svg
            .attr('width', dims.width + dims.margin.left)
            .attr('height', dims.height + dims.margin.top)
            .attr('transform', `translate(${dims.margin.left}, ${dims.margin.top})`)
            .append('g');
        return svg;
    }
    static getIndexOfClosestPoint(x, data) {
        const x0 = x.invert(d3.mouse(d3.event.currentTarget)[0]);
        const bisector = d3.bisector((d) => d.x).left;
        const index = bisector(data, x0, 1);
        const d0 = data[index - 1];
        const d1 = data[index];
        const rangeValueOfFirst = x((d0) ? d0.x : null), rangeValueOfSecond = x((d1) ? d1.x : null), rangeValueOfMousePos = d3.mouse(d3.event.currentTarget)[0], closestIndex = Math.abs(rangeValueOfMousePos - rangeValueOfFirst) > Math.abs(rangeValueOfMousePos - rangeValueOfSecond) ? index : index - 1;
        return closestIndex;
    }
    static hideFocus(svg) {
        svg.selectAll('.point')
            .transition()
            .duration(ChartD3Service.focusLength)
            .attr('r', 1)
            .style('opacity', '0');
        svg.selectAll(`.scrubber`)
            .transition()
            .duration(ChartD3Service.focusLength)
            .style('opacity', '0');
    }
    static showFocus(svg, activeIndex) {
        ChartD3Service.hideFocus(svg);
        svg.selectAll(`.point-${activeIndex}`)
            .transition()
            .duration(ChartD3Service.focusLength)
            .attr('r', 4)
            .style('opacity', '1');
        svg.selectAll(`.scrubber-${activeIndex}`)
            .transition()
            .duration(ChartD3Service.focusLength)
            .style('opacity', '1');
    }
    static calculateLeftPad(dims, axisConfigs, data) {
        const maxYValueLength = Math.max(...data.map(d => d.y)).toString().length;
        this.yPad = maxYValueLength * 4;
        if (axisConfigs.y.showAxis || axisConfigs.y.showTicks) {
            return this.yPad + dims.margin.left;
        }
        else {
            return 0;
        }
    }
    static calculateTicks(dim) {
        return Math.max(dim / 100, 2);
    }
    static formatDataAsStacked(rawData) {
        const stackFactory = d3.stack()
            .offset(ChartD3Service.stackOffsetDiverging)
            .keys(rawData.map((_s, i) => `namey-${i}`));
        const data = ChartD3Service.seriesAsStackFormat(rawData);
        return stackFactory(data);
    }
    static seriesAsStackFormat(series) {
        const namedSeries = series.map((s, i) => {
            return {
                name: `namey-${i}`,
                coords: s
            };
        });
        const collector = [];
        namedSeries.forEach((s) => {
            s.coords.forEach((coords, i) => {
                if (!collector[i]) {
                    collector[i] = {};
                }
                collector[i].categoryValue = coords.x;
                collector[i][s.name] = coords.y;
            });
        });
        return collector;
    }
    static stackOffsetDiverging(series, seriesOrder) {
        const n = series.length;
        if (n < 1) {
            return;
        }
        const firstSeriesIndex = seriesOrder[0];
        for (let i, j = 0, dy, yp, yn, columnCount = series[firstSeriesIndex].length; j < columnCount; ++j) {
            for (yp = yn = 0, i = 0; i < n; ++i) {
                const serie = series[seriesOrder[i]];
                const dataPoint = serie[j];
                dy = dataPoint[1] - dataPoint[0];
                if (dy >= 0) {
                    dataPoint[0] = yp;
                    dataPoint[1] = yp += dy;
                }
                else if (dy < 0) {
                    dataPoint[1] = yn;
                    dataPoint[0] = yn += dy;
                }
                else {
                    dataPoint[0] = yp;
                }
            }
        }
    }
}
ChartD3Service.colors = [
    '#384b5d',
    '#6ec5ab',
    '#6D95A0',
    '#995D81',
    '#586BA4',
    '#58A4B0',
    '#D496A7',
    '#C8AD55',
    '#8EA4D2',
    '#4ECDC4',
    '#98D2EB',
    '#857C8D'
];
ChartD3Service.animationLength = 2500;
ChartD3Service.focusLength = 400;
ChartD3Service.yPad = 15;
ChartD3Service.groupPadding = 2;
